`timescale 1ns/1ns

//To avoid saturation the gain of the filter should not be bigger than 1 in modulus!!!!!
//The coeficients are expressed on 36bits in 2's complement and are considered as multiplied by 2^-35 (1 is not representable while -1 is)

module fir_dec (
	clk,
	clrn,
	vme_cs,
	vme_wr,
	vme_addr,
	vme_data,
	dav_in,
	data_in,
	dav_out,
	data_out,
	overflow);
	
input clk, clrn;

input vme_cs, vme_wr;
input [7:0] vme_addr;
input [31:0] vme_data;

input dav_in;
input [23:0] data_in;
output dav_out;
output [31:0] data_out;
output overflow;

reg dav_out, overflow;
reg [31:0] data_out;

parameter dly= 1'b1;



integer i;

wire [71:0] mult_out;

reg [35:0] fir_coef [31:0]; //coefs for the FIR
reg [23:0] din_shreg [31:1]; //data for the FIR

reg	[71:0] acc_fir; //accumulator

reg [3:0] counter_dec, dec_reg; //dec_reg contains the number of values out of fir1 to reject: writing 9 would mean real "decimation"
reg dav_dec; //signals that a new data out of the decimator is already in the acc_fir accumulator 

reg [4:0] counter_fir;

reg state;

parameter s_idle	= 1'b0;
parameter s_fir		= 1'b1;

parameter addr_reg_space	= 2'b0;
parameter add_fir_space		= 2'b1;

parameter addr_dec_reg	= 6'h0;

wire sel_reg_space, sel_fir_coef_space;

//Address space selection

assign sel_reg_space		= vme_cs && vme_addr[7:6]==addr_reg_space;
assign sel_fir_coef_space	= vme_cs && vme_addr[7:6]==add_fir_space;

//STATE MACHINE

always @(posedge clk or negedge clrn) begin
	if (~clrn) begin
		state	<= # dly s_idle;
	end else begin
		case (state)
		s_idle: begin
			if (dav_in) begin
				state	<= # dly s_fir;
			end				
		end
		s_fir: begin
			if (counter_fir==5'h0) begin
				state	<= # dly s_idle;	
			end
		end
		default: state	<= # dly s_idle;
		endcase
	end
end 

//EXECUTED TAPs' COUNTER

always @(posedge clk or negedge clrn) begin
	if (~clrn) begin
		counter_fir	<= # dly 5'h0;
	end else if (dav_in || |counter_fir) begin
		counter_fir	<= # dly counter_fir + 1'b1;
	end
end

//SHIFT REGISTER FOR DATA IN & DECIMATED FIR DATA

always @(posedge clk) begin
	if (dav_in || |counter_fir) begin
		for (i=1 ; i<32	; i=i+1) begin
			if (i==1) begin
				if (dav_in) begin 
					din_shreg[1]	<= # dly data_in;
				end else if (|counter_fir) begin
			 		din_shreg[1]	<= # dly din_shreg[31];
				end
			end else begin
				din_shreg[i]	<= # dly din_shreg[i-1];	
			end
		end
	end
end	

//FIR COEFICENTS' MEMORY

always @(posedge clk) begin
	if (~clrn) begin
		fir_coef[31]	<= # dly 36'h800000000; //equivalent to -1
		for (i=0 ; i<31	; i=i+1) begin
			fir_coef[i]	<= # dly 36'h0;
		end
	end else if (vme_wr && sel_fir_coef_space) begin
		if (~vme_addr[0]) begin
			fir_coef[31-vme_addr[5:1]][35:4]	<= # dly vme_data;   //to have @ addr i the coef applied to x*Z^-i
		end else begin
			fir_coef[31-vme_addr[5:1]][3:0]	<= # dly vme_data[31:28];   //to have @ addr i the coef applied to x*Z^-i		
		end
	end
end

//DECIMATOR

always @(posedge clk or negedge clrn) begin
	if (~clrn) begin
		dec_reg	<= # dly 4'h0; //no decimation
	end else if (vme_wr && sel_reg_space && vme_addr[5:0]==addr_dec_reg) begin
		dec_reg	<= # dly vme_data[3:0];
	end
end

always @(posedge clk or negedge clrn) begin
	if (~clrn) begin
		counter_dec	<= # dly 4'h0;
		dav_dec		<= # dly 1'b0;
	end else begin
		if (state==s_fir && counter_fir==5'h0) begin
			if (counter_dec==dec_reg) begin
				counter_dec	<= # dly 4'h0;
				dav_dec		<= # dly 1'b1;
			end else begin
				counter_dec	<= # dly counter_dec + 1'b1;
				dav_dec		<= # dly 1'b0;
			end
		end else begin
			dav_dec	<= # dly 1'b0;
		end
	end
end

//MULTIPLICATION AND ACCUMULATION

sign_mult_36x36 sign_mult_i(
	.clock0(clk),
	.dataa_0({{12{din_shreg[31][23]}}, din_shreg[31]}),
	.datab_0(fir_coef[counter_fir]),
	.result(mult_out));

always @(posedge clk or negedge clrn) begin
	if (~clrn) begin
		acc_fir	<= # dly 72'h0;
	end else begin
		if (dav_in) begin
			acc_fir	<= # dly 72'h0;
		end else if (state==s_fir) begin
			acc_fir	<= # dly acc_fir + mult_out;
		end
	end
end

//DATA_OUT DAV_OUT OVERFLOW

reg overexteem;
reg [31:0] rounded;

always@ (posedge clk or negedge clrn) begin
	if (~clrn) begin
		overexteem	<= # dly 1'b0;
	end else if (dav_dec && acc_fir[26] && ~|acc_fir[25:0]) begin
		overexteem	<= # dly ~overexteem;
	end
end	

always @(acc_fir or overexteem) begin
	if (acc_fir[27] && (overexteem || |acc_fir[26:0])) begin
		rounded	= acc_fir[58:27] + 1'b1;
	end else begin
		rounded	= acc_fir[58:27];
	end
end
	
always @(posedge clk or negedge clrn) begin
	if (~clrn) begin
		dav_out		<= # dly 1'b0;
		data_out	<= # dly 32'h0;	 
		overflow	<= # dly 1'b0;	
	end else begin
		if (dav_dec) begin
			dav_out		<= # dly 1'b1;
			if (acc_fir[71]==rounded[31] && acc_fir[58]==rounded[31]) begin				
				data_out	<= # dly rounded; 
				overflow	<= # dly 1'b0; 
			end else begin
				data_out	<= # dly acc_fir[71] ? 32'h8000_0000 : 32'h7fff_ffff;	
				overflow	<= # dly 1'b1;
			end
		end else begin
			dav_out		<= # dly 1'b0;
			overflow	<= # dly 1'b0;
		end			
	end
end
	
endmodule