`timescale 1ns/1ns

module q_pll
	(input	clk,
	input 	clrn,
	
	input 	start,
	input	[31:0] ampl_i,
	input	[31:0] ampl_q,
	output	done,
	output	[23:0] cos_pll,
	output	[23:0] sin_pll,
	
	input	vme_cs,
	input	vme_wr,
	input	vme_rd,
	input	[5:0] vme_addr, 
	input	[31:0] vme_din,
	output	reg vme_dav,
	output	reg	[31:0] vme_dout);

parameter dly = 1;

parameter reg_space_base_addr	= 2'd0;
parameter pid_a_base_addr		= 2'd1;
parameter pid_p_base_addr		= 2'd2;

parameter addr_status		= 4'd0;
parameter addr_ampl_dem_q	= 4'd1;
parameter addr_phy_dem_q	= 4'd2;
parameter addr_ampl_dem_i	= 4'd3;
parameter addr_phy_dem_i	= 4'd4;
parameter addr_des_ampl		= 4'd5;
parameter addr_phy_corr		= 4'd6;
parameter addr_oow_a0_limit	= 4'd7;
parameter addr_oow_a1_limit	= 4'd8;
parameter addr_oow_ampl		= 4'd9;
parameter addr_oow_df		= 4'd10;
parameter addr_oow_f1		= 4'd11;
parameter addr_oow_f2		= 4'd12;
parameter addr_pll_freq		= 4'd13;
parameter addr_pll_ampl		= 4'd14;

reg [39:0] phase_pll;
reg [41:0] ampl_pll;  
wire [35:0] cos_pll_ext, sin_pll_ext;
wire done_iq;
wire [35:0] ampl_dem_i, ampl_dem_q;
wire [39:0] phy_dem_i, phy_dem_q;
reg	[31:0] des_ampl, phy_corr;
wire vme_reg_space_cs, vme_pid_a_cs, vme_pid_b_cs;
wire [35:0] pid_a_input, pid_p_input;
wire [31:0] vme_dout_a_pid, vme_dout_p_pid;
wire vme_dav_a_pid, vme_dav_p_pid;
wire dampl_dav, df_dav;
wire [31:0] dampl, df;
reg [31:0] oow_a0_limit, oow_a1_limit, oow_ampl;
reg oow, oow_old;
reg [15:0] status_reg;
wire disable_oow_ampl, disable_oow_phase, disable_ampl_control;
reg [31:0] freq_pll;
reg [31:0] oow_df, oow_f1, oow_f2;
reg pos_neg_chirp;


//*******************************
// BASE ADDRESSES DEFINITION
//*******************************

assign vme_reg_space_cs	= vme_addr[5:4]==reg_space_base_addr && vme_cs;
assign vme_pid_a_cs		= vme_addr[5:4]==pid_a_base_addr && vme_cs;
assign vme_pid_p_cs		= vme_addr[5:4]==pid_p_base_addr && vme_cs;

//********************************
// cordic mixer - demixer
//********************************

cordic_demodulator demod (
	.clk(clk),
	.clrn(clrn),
	
	.phase_pll(phase_pll),
	.ampl_pll(ampl_pll),
	.ampl_i({{2{ampl_i[31]}} ,ampl_i , 8'b0}),
	.ampl_q({{2{ampl_q[31]}} ,ampl_q , 8'b0}),
	.start(start),
	
	.done_pll(done),
	.cos_out_pll(cos_pll_ext),
	.sin_out_pll(sin_pll_ext),
	
	.done_iq(done_iq),
	.ampl_out_i(ampl_dem_i),
	.phase_out_i(phy_dem_i),
	.ampl_out_q(ampl_dem_q),
	.phase_out_q(phy_dem_q));
	
assign cos_pll	= cos_pll_ext[35:12]; 
assign sin_pll	= sin_pll_ext[35:12]; 

//***********************************
// FEEDBACK LOOPS
//***********************************

always @(posedge clk or negedge clrn) begin
	if (~clrn) begin
		des_ampl	<= # dly 32'h0040_0000;
	end else if (vme_reg_space_cs && vme_wr && vme_addr[3:0]==addr_des_ampl) begin
		des_ampl	<= # dly vme_din;
	end
end

always @(posedge clk or negedge clrn) begin
	if (~clrn) begin
		phy_corr	<= # dly 32'h0000_0000;
	end else if (vme_reg_space_cs && vme_wr && vme_addr[3:0]==addr_phy_corr) begin
		phy_corr	<= # dly vme_din;
	end
end

assign pid_a_input	= ampl_dem_i - {1'b0, des_ampl, 3'b0};
assign pid_p_input	= {phy_corr, 4'b0} - phy_dem_i[39:4];


pll_pid amplitude_pid(
	.clk(clk), 
	.clrn(clrn),
	.vme_cs(vme_pid_a_cs),
	.vme_wr(vme_wr),
	.vme_rd(vme_rd),
	.vme_addr(vme_addr[2:0]), 
	.vme_din(vme_din),
	.vme_dav(vme_dav_a_pid),
	.vme_dout(vme_dout_a_pid),	
	.err_dav(done_iq),
	.err(pid_a_input),
	.df_dav(dampl_dav), 
	.df(dampl));
	
pll_pid phase_pid(
	.clk(clk), 
	.clrn(clrn),
	.vme_cs(vme_pid_p_cs),
	.vme_wr(vme_wr),
	.vme_rd(vme_rd),
	.vme_addr(vme_addr[2:0]), 
	.vme_din(vme_din),
	.vme_dav(vme_dav_p_pid),
	.vme_dout(vme_dout_p_pid),	
	.err_dav(done_iq),
	.err(pid_p_input),
	.df_dav(df_dav), 
	.df(df));

//****************************************
//OUT OF (LOCKING) WINDOW SETTINGS
//****************************************	
		
always @(posedge clk or negedge clrn) begin
	if (~clrn) begin
		oow_a0_limit	<= # dly 32'd1_000;
	end else if (vme_reg_space_cs && vme_wr && vme_addr[3:0]==addr_oow_a0_limit) begin
		oow_a0_limit	<= # dly vme_din;
	end
end

always @(posedge clk or negedge clrn) begin
	if (~clrn) begin
		oow_a1_limit	<= # dly 32'd360_000;
	end else if (vme_reg_space_cs && vme_wr && vme_addr[3:0]==addr_oow_a1_limit) begin
		oow_a1_limit	<= # dly vme_din;
	end
end

always @(posedge clk or negedge clrn) begin
	if (~clrn) begin
		oow_ampl	<= # dly 32'h4000_0000;
	end else if (vme_reg_space_cs && vme_wr && vme_addr[3:0]==addr_oow_ampl) begin
		oow_ampl	<= # dly vme_din;
	end
end

always @(posedge clk or negedge clrn) begin
	if (~clrn) begin
		oow_f1	<= # dly 32'h0600_0000;
	end else if (vme_reg_space_cs && vme_wr && vme_addr[3:0]==addr_oow_f1) begin
		oow_f1	<= # dly vme_din;
	end
end

always @(posedge clk or negedge clrn) begin
	if (~clrn) begin
		oow_f2	<= # dly 32'h1000_0000;
	end else if (vme_reg_space_cs && vme_wr && vme_addr[3:0]==addr_oow_f2) begin
		oow_f2	<= # dly vme_din;
	end
end

always @(posedge clk or negedge clrn) begin
	if (~clrn) begin
		oow_df	<= # dly 31'h0004_0000;
	end else if (vme_reg_space_cs && vme_wr && vme_addr[3:0]==addr_oow_df) begin
		oow_df	<= # dly vme_din[30:0];
	end
end
	
always @(posedge clk or negedge clrn) begin
	if (~clrn) begin
		oow_old	<= # dly 1'b0;
		oow		<= # dly 1'b1;
	end else begin
		oow_old	<= # dly oow;
		if (ampl_dem_i[35:4] < oow_a0_limit) begin
			oow		<= # dly 1'b1;
		end else if (ampl_dem_i[35:4] > oow_a1_limit) begin
			oow		<= # dly 1'b0;
		end
	end
end


//*********************************************
//PLL AMPLITUDE FREQUENCY AND PHASE
//*********************************************

always @(posedge clk or negedge clrn) begin
	if (~clrn) begin
		ampl_pll	<= # dly 42'h10_0000_0000;
	end else if (vme_reg_space_cs && vme_wr && vme_addr[3:0]==addr_pll_ampl) begin
		ampl_pll	<= # dly {1'b0, vme_din, 9'h0};
	end else if (oow && ~disable_oow_ampl) begin
		ampl_pll	<= # dly {1'b0, oow_ampl, 9'h0};
	end else if (dampl_dav && ~disable_ampl_control && dampl_dav) begin
		ampl_pll	<= # dly ampl_pll + {dampl_dav, 10'h0};
	end
end
	
always @(posedge clk or negedge clrn) begin
	if (~clrn) begin
		freq_pll		<= # dly 32'h0010_0000;
		pos_neg_chirp	<= # dly 1'b1;
	end else if (vme_reg_space_cs && vme_wr && vme_addr[3:0]==addr_pll_freq) begin
		freq_pll	<= # dly vme_din;
	end else if (oow && ~oow_old && ~disable_oow_phase) begin
		freq_pll	<= # dly oow_f1;
	end else if (oow && ~disable_oow_phase && df_dav) begin
		if (freq_pll>oow_f2) begin
			pos_neg_chirp	<= # dly 1'b0;
			freq_pll		<= # dly freq_pll - {1'b0,oow_df};
		end else if (freq_pll<oow_f1) begin
			pos_neg_chirp	<= # dly 1'b1;
			freq_pll		<= # dly freq_pll + {1'b0,oow_df};
		end else begin
			freq_pll	<= # dly pos_neg_chirp ? freq_pll + {1'b0,oow_df}: freq_pll - {1'b0,oow_df};
		end
	end else begin
		freq_pll	<= # dly freq_pll + df;
	end
end

always @(posedge clk or negedge clrn) begin
	if (~clrn) begin
		phase_pll	<= # dly 40'h0;
	end else begin
		phase_pll	<= # dly phase_pll + freq_pll;
	end
end

//***********************************************
//STATUS/CONTROL REGISTER
//***********************************************

always @(posedge clk or negedge clrn) begin
	if (~clrn) begin
		status_reg	<= # dly 16'h0006;
	end else if (vme_reg_space_cs && vme_wr && vme_addr[3:0]==addr_phy_corr) begin
		status_reg	<= # dly vme_din[15:0];
	end
end

assign disable_ampl_control	= status_reg[2];
assign disable_oow_ampl		= status_reg[1];
assign disable_oow_phase	= status_reg[0];

//**********************************************
//VME READOUT
//**********************************************

always @* begin
	if (vme_reg_space_cs) begin
		vme_dav	= vme_rd;
		case (vme_addr[3:0])
			addr_status			:	vme_dout	= {15'h0, oow, status_reg};
			addr_ampl_dem_q		:	vme_dout	= ampl_dem_q[34:3];  
			addr_phy_dem_q		:	vme_dout	= phy_dem_q[39:8];   
			addr_ampl_dem_i		:	vme_dout	= ampl_dem_i[34:3];  
			addr_phy_dem_i		:	vme_dout	= phy_dem_i[39:8];   
			addr_des_ampl		:	vme_dout	= des_ampl;	   
			addr_phy_corr		:	vme_dout	= phy_corr;	   
			addr_oow_a0_limit	:	vme_dout	= oow_a0_limit;
			addr_oow_a1_limit	:   vme_dout	= oow_a1_limit;
			addr_oow_ampl		:   vme_dout	= oow_ampl;
			addr_oow_df		    :   vme_dout	= oow_df;
			addr_oow_f1		    :   vme_dout	= oow_f1;
		    addr_oow_f2		    :   vme_dout	= oow_f2;
		   	addr_pll_freq		:   vme_dout	= freq_pll;
            addr_pll_ampl		:   vme_dout	= ampl_pll[40:9];
            default				:   vme_dout	= 32'hdead_beef;
		endcase
	end else if (vme_pid_a_cs) begin
		vme_dav		= vme_dav_a_pid;
		vme_dout	= vme_dout_a_pid;
	end else if (vme_pid_p_cs) begin
		vme_dav		= vme_dav_a_pid;
		vme_dout	= vme_dout_a_pid;
	end else begin
		vme_dav		= 1'b0;
		vme_dout	= 32'hdead_beef;
	end
end		

endmodule