/****************************************************************************************
*
*    File Name: as7c33512pfs32.v 
*      Version:  2.0
*         Date:  APRIL 15th,2005
*        Model:  BUS Functional
*    Simulator:  Cadence Verilog-XL
*
*
*      Company:  Alliance Semiconductor pvt ltd.
*  Part Number:  as7c33512pfs32 (512K x 32)
*
*  Description:  Alliance 18Mb Syncburst SRAM (Pipelined SCD)
*
*
* Rev   Changes
* ---   ------------------------------------------------------
* 2.0    - tCD (access time) change according to new datasheet & Ft_n pin removed
*                
****************************************************************************************/

// DO NOT CHANGE THE TIMESCALE
// MAKE SURE YOUR SIMULATOR USE "PS" RESOLUTION
`timescale 1ns / 10ps

//`define FREQ166

module as7c33512pfs32 (dq, addr, lbo_n, adv_n, clk, adsc_n, adsp_n, bwa_n, bwb_n, bwc_n, bwd_n, bwe_n, gw_n, ce0_n, ce1, ce2_n, oe_n, zz);

    // Constant Parameters
    parameter                       addr_bits =     19;         //  19 bits
    parameter                       data_bits =     32;         //  32 bits
    parameter                       mem_sizes =  524288;        //  512K

    // Timing Parameters 
    `ifdef FREQ133
     parameter                       tHZC      =     3.8;
     
    `else 
    //`ifdef FREQ166
     parameter                       tHZC      =     3.4;
    
   //`endif
    `endif

    // Ports Declaration
    inout     [31 : 0] dq;                         
    input     [31 : 0] addr;                       
    input                           lbo_n;                      
    input                           adv_n;                      
    input                           clk;                        
    input                           adsc_n;                     
    input                           adsp_n;                     
    input                           bwa_n;                      
    input                           bwb_n;                      
    input                           bwc_n;                      
    input                           bwd_n;                      
    input                           bwe_n;                      
    input                           gw_n;                       
    input                           ce0_n;                      
    input                           ce1;                        
    input                           ce2_n;                      
    input                           oe_n;                       
    input                           zz;                         

    // Memory Array
    reg [((data_bits / 4) - 1) : 0] bank0 [0 : mem_sizes-1];      // Memory Bank 0
    reg [((data_bits / 4) - 1) : 0] bank1 [0 : mem_sizes-1];      // Memory Bank 1
    reg [((data_bits / 4) - 1) : 0] bank2 [0 : mem_sizes-1];      // Memory Bank 2
    reg [((data_bits / 4) - 1) : 0] bank3 [0 : mem_sizes-1];      // Memory Bank 3

    reg       [(data_bits - 1) : 0] din;                        // Input Registers
    reg       [(data_bits - 1) : 0] dout;                       // Output Registers

    reg       [(addr_bits - 1) : 0] addr_reg_in;                // address Register In
    reg       [(addr_bits - 1) : 0] addr_reg_out;                // address Register out
    reg       [(addr_bits - 1) : 0] addr_reg_read;              // address Register for Read Operation
    reg       [(addr_bits - 1) : 0] addr_reg_write;             // address Register for Write Operation
    reg                             baddr1;                     // Burst address 1
    reg                             baddr0;                     // Burst address 0
    reg                     [1 : 0] bcount;                     // 2-bit Burst Counter

    reg                             ce_reg;
    reg                             pipe_reg;
    reg                             bwa_reg;
    reg                             bwb_reg;
    reg                             bwc_reg;
    reg                             bwd_reg;
    reg                             sys_clk;

    wire                            ce      = (~ce0_n & ce1 & ~ce2_n);
    wire                            bwa_n_int   = ((bwa_n | bwe_n) & gw_n | (~ce0_n & ~adsp_n));
    wire                            bwb_n_int   = ((bwb_n | bwe_n) & gw_n | (~ce0_n & ~adsp_n));
    wire                            bwc_n_int   = ((bwc_n | bwe_n) & gw_n | (~ce0_n & ~adsp_n));
    wire                            bwd_n_int   = ((bwd_n | bwe_n) & gw_n | (~ce0_n & ~adsp_n));
    wire                            clr     = (~adsc_n | (~adsp_n & ~ce0_n));
    wire                            Ft_n    = 1'b1;

    // Initialize
    initial begin
        ce_reg = 1'b0;
        sys_clk = 1'b0;
        pipe_reg = 1'b0;
        $timeformat (-9, 1, " ns", 10);                         // Format time unit
    end

    // System Clock
    always begin
        @ (posedge clk) begin
            sys_clk = ~zz;
        end
        @ (negedge clk) begin
            sys_clk = 1'b0;
        end
    end

    always @ (posedge sys_clk) begin
        // address Register
        if (clr) addr_reg_in = addr;

        // Binary Counter and Logic
        if      ( lbo_n  &  clr) bcount = 0;                    // Interleaved Burst
        else if (~lbo_n  &  clr) bcount = addr [1 : 0];         // Linear Burst
        else if (~adv_n & ~clr) bcount = (bcount + 1);         // Advance Counter

        // Burst address Decode
        assign baddr1 = lbo_n ? (bcount [1] ^ addr_reg_in [1]) : bcount [1];
        assign baddr0 = lbo_n ? (bcount [0] ^ addr_reg_in [0]) : bcount [0];

        // Pipeline Read address - After one cycle 
        addr_reg_read = addr_reg_out;

        // Output address
        addr_reg_out = {addr_reg_in [(addr_bits - 1) : 2], baddr1, baddr0};

        
	// Write address - same cycle for pipeline & Flow-through
	addr_reg_write = addr_reg_out;
	
	
	// flowthrough read address - same cycle
	if(~Ft_n)
	addr_reg_read = addr_reg_out;
	
	
	// Byte Write Register    
        bwa_reg = ~bwa_n_int;
        bwb_reg = ~bwb_n_int;
        bwc_reg = ~bwc_n_int;
        bwd_reg = ~bwd_n_int;

        // Pipelined Enable
        pipe_reg = ce_reg;

        // Enable Register
        if (clr) ce_reg = ce;

        // Input Register
        if (ce_reg & (bwa_reg | bwb_reg | bwc_reg | bwd_reg)) begin
            din = dq;
        end

        // Byte Write Driver
        if (ce_reg & bwa_reg) begin
            bank0 [addr_reg_write] = din [ 7 :  0];
        end
        if (ce_reg & bwb_reg) begin
            bank1 [addr_reg_write] = din [15 :  8];
        end
        if (ce_reg & bwc_reg) begin
            bank2 [addr_reg_write] = din [23 : 16];
        end
        if (ce_reg & bwd_reg) begin
            bank3 [addr_reg_write] = din [31 : 24];
        end

        // Output Register
        if (~(bwa_reg | bwb_reg | bwc_reg | bwd_reg)) begin
            dout [ 7 :  0] = bank0 [addr_reg_read];
            dout [15 :  8] = bank1 [addr_reg_read];
            dout [23 : 16] = bank2 [addr_reg_read];
            dout [31 : 24] = bank3 [addr_reg_read];
        end
    end

    // Output Buffer
        
    assign #tHZC dq = (Ft_n) ? ((~oe_n & ~zz  & ce_reg & pipe_reg & ~(bwa_reg | bwb_reg | bwc_reg | bwd_reg)) ? dout : 32'bz) :                           
                                ((~oe_n & ~zz & ce_reg   & ~(bwa_reg | bwb_reg | bwc_reg | bwd_reg)) ? dout : 32'bz);
    

    // Timing Check 
    specify
        specparam   
	
	
	`ifdef FREQ133
	            tCYC    =  7.5,     // Clock        - Clock cycle time
                    tCH     =  2.4,     //                Clock HIGH time
                    tCL     =  2.4,     //                Clock LOW time

                    tAS     =  1.5,     // Setup Times  - address
                    tADSCS  =  1.5,     //                address Status
		    tADSPS  =  1.5,     //                address Status
                    tADVS   =  1.5,     //                address Advance
                    tWS     =  1.5,     //                Byte Write Enables
                    tDS     =  1.5,     //                Data-in
                    tCSS    =  1.5,     //                Chip Enable
                    
                    tAH     =  0.5,     // Hold Times   - address                                  
                    tADSCH  =  0.5,     //                address Status
		    tADSPH  =  0.5,     //                address Status
                    tADVH   =  0.5,     //                address Advance
                    tWH     =  0.5,     //                Byte Write Enables
                    tDH     =  0.5,     //                Data-in                                  
                    tCSH    =  0.5;     //                Chip Enable
		    
	`else 
	//`ifdef FREQ166
	            tCYC    =  6.0,     // Clock        - Clock cycle time
                    tCH     =  2.4,     //                Clock HIGH time
                    tCL     =  2.3,     //                Clock LOW time

                    tAS     =  1.5,     // Setup Times  - address
                    tADSCS  =  1.5,     //                address Status
		    tADSPS  =  1.5,     //                address Status
                    tADVS   =  1.5,     //                address Advance
                    tWS     =  1.5,     //                Byte Write Enables
                    tDS     =  1.5,     //                Data-in
                    tCSS    =  1.5,     //                Chip Enable
                    
                    tAH     =  0.5,     // Hold Times   - address                                  
                    tADSCH  =  0.5,     //                address Status
		    tADSPH  =  0.5,     //                address Status
                    tADVH   =  0.5,     //                address Advance
                    tWH     =  0.5,     //                Byte Write Enables
                    tDH     =  0.5,     //                Data-in                                  
                    tCSH    =  0.5;     //                Chip Enable	
        
	//`endif
	`endif
	

        $width      (negedge clk, tCL);
        $width      (posedge clk, tCH);
        $period     (negedge clk, tCYC);
        $period     (posedge clk, tCYC);
        $setuphold  (posedge clk, adsp_n, tADSPS, tADSPH);
        $setuphold  (posedge clk, adsc_n, tADSCS, tADSCH);
        $setuphold  (posedge clk, addr,   tAS,   tAH);
        $setuphold  (posedge clk, bwa_n,  tWS,   tWH);
        $setuphold  (posedge clk, bwb_n,  tWS,   tWH);
        $setuphold  (posedge clk, bwc_n,  tWS,   tWH);
        $setuphold  (posedge clk, bwd_n,  tWS,   tWH);
        $setuphold  (posedge clk, bwe_n,  tWS,   tWH);
        $setuphold  (posedge clk, gw_n,   tWS,   tWH);
        $setuphold  (posedge clk, ce0_n,   tCSS,  tCSH);
        $setuphold  (posedge clk, ce1,    tCSS,  tCSH);
        $setuphold  (posedge clk, ce2_n,  tCSS,  tCSH);
        $setuphold  (posedge clk, adv_n,  tADVS,  tADVH);
    endspecify                        

endmodule

