/****************************************************************************************
*
*    File Name: as7c33512pfs32.v 
*      Version:  2.0
*         Date:  APRIL 15th,2005
*        Model:  BUS Functional
*    Simulator:  Cadence Verilog-XL
*
*
*      Company:  Alliance Semiconductor pvt ltd.
*  Part Number:  as7c33512pfs32 (512K x 32)
*
*  Description:  Alliance 18Mb Syncburst SRAM (Pipelined SCD)
*
*
* Rev   Changes
* ---   ------------------------------------------------------
* 2.0    - tCD (access time) change according to new datasheet & Ft_n pin removed
*                
****************************************************************************************/

// DO NOT CHANGE THE TIMESCALE
// MAKE SURE YOUR SIMULATOR USE "PS" RESOLUTION
`timescale 1ns / 10ps

//`define FREQ166

module as7c33512pfs32 (Dq, Addr, Lbo_n, Adv_n, Clk, Adsc_n, Adsp_n, Bwa_n, Bwb_n, Bwc_n, Bwd_n, Bwe_n, Gw_n, Ce0_n, Ce1, Ce2_n, Oe_n, Zz);

    // Constant Parameters
    parameter                       addr_bits =     19;         //  19 bits
    parameter                       data_bits =     32;         //  32 bits
    parameter                       mem_sizes =  524288;        //  512K

    // Timing Parameters 
    `ifdef FREQ133
     parameter                       tHZC      =     3.8;
     
    `else 
    //`ifdef FREQ166
     parameter                       tHZC      =     3.4;
    
   //`endif
    `endif

    // Ports Declaration
    inout     [(data_bits - 1) : 0] Dq;                         
    input     [(addr_bits - 1) : 0] Addr;                       
    input                           Lbo_n;                      
    input                           Adv_n;                      
    input                           Clk;                        
    input                           Adsc_n;                     
    input                           Adsp_n;                     
    input                           Bwa_n;                      
    input                           Bwb_n;                      
    input                           Bwc_n;                      
    input                           Bwd_n;                      
    input                           Bwe_n;                      
    input                           Gw_n;                       
    input                           Ce0_n;                      
    input                           Ce1;                        
    input                           Ce2_n;                      
    input                           Oe_n;                       
    input                           Zz;                         

    // Memory Array
    reg [((data_bits / 4) - 1) : 0] bank0 [0 : mem_sizes-1];      // Memory Bank 0
    reg [((data_bits / 4) - 1) : 0] bank1 [0 : mem_sizes-1];      // Memory Bank 1
    reg [((data_bits / 4) - 1) : 0] bank2 [0 : mem_sizes-1];      // Memory Bank 2
    reg [((data_bits / 4) - 1) : 0] bank3 [0 : mem_sizes-1];      // Memory Bank 3

    reg       [(data_bits - 1) : 0] din;                        // Input Registers
    reg       [(data_bits - 1) : 0] dout;                       // Output Registers

    reg       [(addr_bits - 1) : 0] addr_reg_in;                // Address Register In
    reg       [(addr_bits - 1) : 0] addr_reg_out;                // Address Register out
    reg       [(addr_bits - 1) : 0] addr_reg_read;              // Address Register for Read Operation
    reg       [(addr_bits - 1) : 0] addr_reg_write;             // Address Register for Write Operation
    reg                             baddr1;                     // Burst Address 1
    reg                             baddr0;                     // Burst Address 0
    reg                     [1 : 0] bcount;                     // 2-bit Burst Counter

    reg                             ce_reg;
    reg                             pipe_reg;
    reg                             bwa_reg;
    reg                             bwb_reg;
    reg                             bwc_reg;
    reg                             bwd_reg;
    reg                             sys_clk;

    wire                            ce      = (~Ce0_n & Ce1 & ~Ce2_n);
    wire                            bwa_n   = ((Bwa_n | Bwe_n) & Gw_n | (~Ce0_n & ~Adsp_n));
    wire                            bwb_n   = ((Bwb_n | Bwe_n) & Gw_n | (~Ce0_n & ~Adsp_n));
    wire                            bwc_n   = ((Bwc_n | Bwe_n) & Gw_n | (~Ce0_n & ~Adsp_n));
    wire                            bwd_n   = ((Bwd_n | Bwe_n) & Gw_n | (~Ce0_n & ~Adsp_n));
    wire                            clr     = (~Adsc_n | (~Adsp_n & ~Ce0_n));
    wire                            Ft_n    = 1'b1;

    // Initialize
    initial begin
        ce_reg = 1'b0;
        sys_clk = 1'b0;
        pipe_reg = 1'b0;
        $timeformat (-9, 1, " ns", 10);                         // Format time unit
    end

    // System Clock
    always begin
        @ (posedge Clk) begin
            sys_clk = ~Zz;
        end
        @ (negedge Clk) begin
            sys_clk = 1'b0;
        end
    end

    always @ (posedge sys_clk) begin
        // Address Register
        if (clr) addr_reg_in = Addr;

        // Binary Counter and Logic
        if      ( Lbo_n  &  clr) bcount = 0;                    // Interleaved Burst
        else if (~Lbo_n  &  clr) bcount = Addr [1 : 0];         // Linear Burst
        else if (~Adv_n & ~clr) bcount = (bcount + 1);         // Advance Counter

        // Burst Address Decode
        assign baddr1 = Lbo_n ? (bcount [1] ^ addr_reg_in [1]) : bcount [1];
        assign baddr0 = Lbo_n ? (bcount [0] ^ addr_reg_in [0]) : bcount [0];

        // Pipeline Read Address - After one cycle 
        addr_reg_read = addr_reg_out;

        // Output Address
        addr_reg_out = {addr_reg_in [(addr_bits - 1) : 2], baddr1, baddr0};

        
	// Write Address - same cycle for pipeline & Flow-through
	addr_reg_write = addr_reg_out;
	
	
	// flowthrough read address - same cycle
	if(~Ft_n)
	addr_reg_read = addr_reg_out;
	
	
	// Byte Write Register    
        bwa_reg = ~bwa_n;
        bwb_reg = ~bwb_n;
        bwc_reg = ~bwc_n;
        bwd_reg = ~bwd_n;

        // Pipelined Enable
        pipe_reg = ce_reg;

        // Enable Register
        if (clr) ce_reg = ce;

        // Input Register
        if (ce_reg & (bwa_reg | bwb_reg | bwc_reg | bwd_reg)) begin
            din = Dq;
        end

        // Byte Write Driver
        if (ce_reg & bwa_reg) begin
            bank0 [addr_reg_write] = din [ 7 :  0];
        end
        if (ce_reg & bwb_reg) begin
            bank1 [addr_reg_write] = din [15 :  8];
        end
        if (ce_reg & bwc_reg) begin
            bank2 [addr_reg_write] = din [23 : 16];
        end
        if (ce_reg & bwd_reg) begin
            bank3 [addr_reg_write] = din [31 : 24];
        end

        // Output Register
        if (~(bwa_reg | bwb_reg | bwc_reg | bwd_reg)) begin
            dout [ 7 :  0] = bank0 [addr_reg_read];
            dout [15 :  8] = bank1 [addr_reg_read];
            dout [23 : 16] = bank2 [addr_reg_read];
            dout [31 : 24] = bank3 [addr_reg_read];
        end
    end

    // Output Buffer
        
    assign #tHZC Dq = (Ft_n) ? ((~Oe_n & ~Zz  & ce_reg & pipe_reg & ~(bwa_reg | bwb_reg | bwc_reg | bwd_reg)) ? dout : 32'bz) :                           
                                ((~Oe_n & ~Zz & ce_reg   & ~(bwa_reg | bwb_reg | bwc_reg | bwd_reg)) ? dout : 32'bz);
    

    // Timing Check 
    specify
        specparam   
	
	
	`ifdef FREQ133
	            tCYC    =  7.5,     // Clock        - Clock cycle time
                    tCH     =  2.4,     //                Clock HIGH time
                    tCL     =  2.4,     //                Clock LOW time

                    tAS     =  1.5,     // Setup Times  - Address
                    tADSCS  =  1.5,     //                Address Status
		    tADSPS  =  1.5,     //                Address Status
                    tADVS   =  1.5,     //                Address Advance
                    tWS     =  1.5,     //                Byte Write Enables
                    tDS     =  1.5,     //                Data-in
                    tCSS    =  1.5,     //                Chip Enable
                    
                    tAH     =  0.5,     // Hold Times   - Address                                  
                    tADSCH  =  0.5,     //                Address Status
		    tADSPH  =  0.5,     //                Address Status
                    tADVH   =  0.5,     //                Address Advance
                    tWH     =  0.5,     //                Byte Write Enables
                    tDH     =  0.5,     //                Data-in                                  
                    tCSH    =  0.5;     //                Chip Enable
		    
	`else 
	//`ifdef FREQ166
	            tCYC    =  6.0,     // Clock        - Clock cycle time
                    tCH     =  2.4,     //                Clock HIGH time
                    tCL     =  2.3,     //                Clock LOW time

                    tAS     =  1.5,     // Setup Times  - Address
                    tADSCS  =  1.5,     //                Address Status
		    tADSPS  =  1.5,     //                Address Status
                    tADVS   =  1.5,     //                Address Advance
                    tWS     =  1.5,     //                Byte Write Enables
                    tDS     =  1.5,     //                Data-in
                    tCSS    =  1.5,     //                Chip Enable
                    
                    tAH     =  0.5,     // Hold Times   - Address                                  
                    tADSCH  =  0.5,     //                Address Status
		    tADSPH  =  0.5,     //                Address Status
                    tADVH   =  0.5,     //                Address Advance
                    tWH     =  0.5,     //                Byte Write Enables
                    tDH     =  0.5,     //                Data-in                                  
                    tCSH    =  0.5;     //                Chip Enable	
        
	//`endif
	`endif
	

        $width      (negedge Clk, tCL);
        $width      (posedge Clk, tCH);
        $period     (negedge Clk, tCYC);
        $period     (posedge Clk, tCYC);
        $setuphold  (posedge Clk, Adsp_n, tADSPS, tADSPH);
        $setuphold  (posedge Clk, Adsc_n, tADSCS, tADSCH);
        $setuphold  (posedge Clk, Addr,   tAS,   tAH);
        $setuphold  (posedge Clk, Bwa_n,  tWS,   tWH);
        $setuphold  (posedge Clk, Bwb_n,  tWS,   tWH);
        $setuphold  (posedge Clk, Bwc_n,  tWS,   tWH);
        $setuphold  (posedge Clk, Bwd_n,  tWS,   tWH);
        $setuphold  (posedge Clk, Bwe_n,  tWS,   tWH);
        $setuphold  (posedge Clk, Gw_n,   tWS,   tWH);
        $setuphold  (posedge Clk, Ce0_n,   tCSS,  tCSH);
        $setuphold  (posedge Clk, Ce1,    tCSS,  tCSH);
        $setuphold  (posedge Clk, Ce2_n,  tCSS,  tCSH);
        $setuphold  (posedge Clk, Adv_n,  tADVS,  tADVH);
    endspecify                        

endmodule

